import copy
import os
from pprint import pprint

import simplejson as json
from assertpy import assert_that
from bson.objectid import ObjectId
from ereuse_devicehub.flaskapp import DeviceHub
from ereuse_devicehub.resources.submitter.grd_submitter.grd_submitter import GRDSubmitter
from ereuse_devicehub.resources.submitter.submitter_caller import SubmitterCaller
from ereuse_devicehub.utils import Naming
from eve.methods.common import parse
from eve.tests import TestMinimal
from flask.ext.pymongo import MongoClient
from passlib.handlers.sha2_crypt import sha256_crypt


class TestBase(TestMinimal):
    DEVICES = 'devices'
    DEVICE_EVENT = 'events/devices'
    EVENTS = 'events'
    PLACES = 'places'
    SNAPSHOT = 'snapshot'
    ACCOUNTS = 'accounts'

    def setUp(self, settings_file=None, url_converters=None):
        from ereuse_devicehub import default_settings as settings
        self.set_settings(settings)
        self.app = DeviceHub()
        self.prepare()

    def set_settings(self, settings):
        settings.MONGO_DBNAME = 'devicehubtest'
        settings.DATABASES = 'dht1', 'dht2'
        settings.DHT1_DBNAME = 'dht1_'
        settings.DHT2_DBNAME = 'dht2_'
        settings.GRD_DEBUG = True  # We do not want to actually fulfill GRD
        settings.APP_NAME = 'DeviceHub'
        settings.DEBUG = True
        settings.TESTING = True
        settings.LOG = True
        settings.GRD = True
        settings.BASE_PATH_SHOWN_TO_GRD = 'https://www.example.com'

    def prepare(self):
        self.MONGO_DBNAME = self.app.config['MONGO_DBNAME']
        self.MONGO_HOST = self.app.config['MONGO_HOST']
        self.MONGO_PORT = self.app.config['MONGO_PORT']
        self.DATABASES = self.app.config['DATABASES']

        self.connection = None
        self.setupDB()

        self.test_client = self.app.test_client()
        self.domain = self.app.config['DOMAIN']

        self.token = self._login()
        self.auth_header = ('authorization', 'Basic ' + self.token)

    def setupDB(self):
        self.connection = MongoClient(self.MONGO_HOST, self.MONGO_PORT)
        self.db = self.connection[self.MONGO_DBNAME]
        self.drop_databases()
        self.create_dummy_user()
        # We call the method again as we have erased the DB
        self.app.grd_submitter_caller = SubmitterCaller(self.app, GRDSubmitter)
        # self.app.grd_submitter_caller.token = self.app.grd_submitter_caller.prepare_user(self.app)
        # self.app.grd_submitter_caller.process = None

    def create_dummy_user(self):
        self.db.accounts.insert(
            {
                'email': "a@a.a",
                'password': sha256_crypt.encrypt('1234'),
                'role': 'superuser',
                'token': 'NOFATDNNUB',
                'databases': self.app.config['DATABASES'],
                'defaultDatabase': self.app.config['DATABASES'][0],
                '@type': 'Account'
            }
        )
        self.account = self.db.accounts.find_one({'email': 'a@a.a'})

    def tearDown(self):
        self.dropDB()
        del self.app

    def drop_databases(self):
        self.connection.drop_database(self.MONGO_DBNAME)
        for database in self.DATABASES:
            self.connection.drop_database(self.app.config[database.upper().replace('-', '') + '_DBNAME'])

    def dropDB(self):
        self.drop_databases()
        self.connection.close()

    def full(self, resource_name: str, resource: dict or str or ObjectId) -> dict:
        return resource if type(resource) is dict else self.get(resource_name, '', str(resource))[0]

    def select_database(self, url):
        if 'accounts' in url:
            return ''
        else:
            return self.app.config['DATABASES'][0]

    def get(self, resource, query='', item=None, authorize=True):
        if resource in self.domain:
            resource = self.domain[resource]['url']
        if item:
            request = '/%s/%s%s' % (resource, item, query)
        else:
            request = '/%s%s' % (resource, query)
        environ_base = {'HTTP_AUTHORIZATION': 'Basic ' + self.token} if authorize else {}
        r = self.test_client.get(self.select_database(resource) + request, environ_base=environ_base)
        return self.parse_response(r)

    def post(self, url, data, headers=None, content_type='application/json'):
        full_url = self.select_database(url) + '/' + url
        if headers is None:
            headers = []
        full_headers = headers + [self.auth_header]
        if type(data) is str:
            # todo this is part of super.post, the only modification is that it does not json.dumps()
            full_headers.append(('Content-Type', content_type))
            r = self.test_client.post(full_url, data=data, headers=full_headers)
            return self.parse_response(r)
        else:
            return super(TestBase, self).post(full_url, data, full_headers, content_type)

    def patch(self, url, data, headers=None):
        if headers is None:
            headers = []
        return super(TestBase, self).patch(self.select_database(url) + '/' + url, data, headers + [self.auth_header])

    def put(self, url, data, headers=None):
        if headers is None:
            headers = []
        return super(TestBase, self).put(self.select_database(url) + '/' + url, data, headers + [self.auth_header])

    def delete(self, url, headers=None):
        if headers is None:
            headers = []
        return super(TestBase, self).delete(self.select_database(url) + '/' + url, headers + [self.auth_header])

    def _login(self) -> str:
        return super(TestBase, self).post('/login', {"email": "a@a.a", "password": "1234"})[0]['token']

    def assert308(self, status):
        self.assertEqual(status, 308)


class TestStandard(TestBase):
    @staticmethod
    def get_json_from_file(filename: str, directory: str = None, parse_json=True, mode='r') -> dict:
        """

        :type filename: str
        :param directory: Optionall. Directory to get the file from. If nothing, it is taken from the actual directory.
        :return:
        """
        if directory is None:
            directory = os.path.dirname(os.path.realpath(__file__))
        with open(os.path.abspath(os.path.join(directory, filename)), mode=mode) as data_file:
            value = json.load(data_file) if parse_json else data_file.read()
        return value

    def parse_event(self, event):
        with self.app.app_context():
            event = parse(event, Naming.resource(event['@type']))
            if 'components' in event:
                for device in event['components'] + [event['device']]:
                    device.update(self.parse_device(device))
            return event

    def parse_device(self, device):
        """
        Parses the device using the standard way of parsing input data, using the settings of DeviceHub.

        This is needed when comparing a fixture with the device it represents, when this comparison happens
        inside with ... app_context()
        :param device:
        :return:
        """
        with self.app.app_context():
            return parse(copy.deepcopy(device), Naming.resource(device['@type']))

    @staticmethod
    def isType(t: str, item: dict):
        return item['@type'] == t

    def assertType(self, t: str, item: dict):
        self.assertEqual(t, item['@type'])

    def assertLen(self, list_to_assert: list, length: int):
        self.assertEqual(len(list_to_assert), length)

    def get_fixture(self, resource_name, file_name, parse_json=True, directory=None, extension='json', mode='r'):
        return self.get_json_from_file('fixtures/{}/{}.{}'.format(resource_name, file_name, extension), directory,
                                       parse_json, mode)

    def post_fixture(self, resource_name, url, file_name):
        return self.post_and_check(url, self.get_fixture(resource_name, file_name))

    def get_first(self, resource_name):
        return self.get_n(resource_name, 0)

    def get_n(self, resource_name, num):
        resources = self.get(resource_name)
        return resources[0]['_items'][num]

    def get_list(self, resource_name: str, identifiers: list):
        return [self.get(resource_name, '', identifier)[0] for identifier in identifiers]

    def post_and_check(self, url, payload):
        response, status_code = self.post(url, payload)
        try:
            self.assert201(status_code)
        except AssertionError as e:
            pprint(response)
            pprint(payload)
            e.message = response
            raise e
        return response

    def patch_and_check(self, url, payload):
        response, status_code = self.patch(url, payload)
        self.assert200(status_code)
        return response

    def put_and_check(self, url, payload):
        response, status_code = self.put(url, payload)
        self.assert200(status_code)
        return response

    def delete_and_check(self, url):
        response, status_code = self.delete(url)
        self.assert204(status_code)
        return response

    def get_fixtures_computers(self) -> list:
        """
        Snapshots four computers in the database, and returns a list with their identifiers.

        One computer has no HID, and it has been Snapshotted with the option 'force_creation' to True.
        :return:
        """
        vaio = self.post_fixture(self.SNAPSHOT, '{}/{}'.format(self.DEVICE_EVENT, self.SNAPSHOT), 'vaio')
        vostro = self.post_fixture(self.SNAPSHOT, '{}/{}'.format(self.DEVICE_EVENT, self.SNAPSHOT), 'vostro')
        xps13 = self.post_fixture(self.SNAPSHOT, '{}/{}'.format(self.DEVICE_EVENT, self.SNAPSHOT), 'xps13')
        mounted = self.get_fixture(self.SNAPSHOT, 'mounted')
        mounted['device']['forceCreation'] = True
        mounted = self.post_and_check('{}/{}'.format(self.DEVICE_EVENT, self.SNAPSHOT), mounted)
        return [self.get(self.EVENTS, '', event['events'][0])[0]['device'] for event in [vaio, vostro, xps13, mounted]]

    def device_and_place_contain_each_other(self, device_id: str, place_id: str) -> list:
        """
        Checks that the materialization of device-place is correct. This is, the place has a reference to a device
        and the device has a reference to a place. If the device has components, this checks the same for the components.
        :param device_id:
        :param place_id:
        :return:
        """
        place, _ = self.get(self.PLACES, '', place_id)
        self.assertIn('devices', place)
        self.assertIn(device_id, place['devices'])
        device, _ = self.get(self.DEVICES, '', device_id)
        self.assertIn('place', device)
        self.assertEqual(place_id, device['place'])
        for component_id in device.get('components', []):
            component, _ = self.get(self.DEVICES, '', component_id)
            self.assertIn('place', component)
            self.assertEqual(component['place'], place_id)

    def devices_do_not_contain_places(self, device_id: str) -> list:
        """ The opposite of `device_and_place_contain_each_other`."""
        device, _ = self.get(self.DEVICES, '', device_id)
        assert_that(device).does_not_contain('place')
        for component_id in device.get('components', []):
            component, _ = self.get(self.DEVICES, '', component_id)
            assert_that(component).does_not_contain('place')



from eve.auth import TokenAuth
from flask import g, current_app

from ereuse_devicehub.exceptions import UnauthorizedToUseDatabase
from ereuse_devicehub.resources.account.domain import AccountDomain, NotADatabase
from ereuse_devicehub.resources.account.role import Role


class RolesAuth(TokenAuth):
    def authorized(self, allowed_roles, resource, method):
        authorized = super(RolesAuth, self).authorized(allowed_roles, resource, method)
        if not authorized and method == 'GET':
            from ereuse_devicehub.resources.device.schema import Device
            if resource == 'devices' or resource in Device.resource_types:
                # We will check if the device is authorized in a hook, later
                # We avoid requesting the device at the database twice
                authorized = True
                self.set_needs_to_be_public()
                self._set_database(True)
            elif resource is None:  # schema (however will only load 'devices')
                authorized = True
        return authorized

    def check_auth(self, token, allowed_roles, resource, method):
        """
        :param token:
        :param allowed_roles: List of JUST ONE Role
        :param resource:
        :param method:
        """

        # use Eve's own db driver; no additional connections/resources are used
        has_perm = False
        if AccountDomain.actual:
            self._set_database()
            if allowed_roles and AccountDomain.actual['role'] >= allowed_roles[0]:
                if not AccountDomain.actual['role'].is_manager():
                    self.set_request_auth_value(AccountDomain.actual['_id'])
                has_perm = True
            elif not allowed_roles:
                has_perm = True
        return has_perm

    def _set_database(self, force: bool = False):
        """
        Sets the actual database according of the value in the URL.
        :param force: Sets the database even if the user has no access to it
        :raises NotADatabase: When the URL is malformed or doesn't contain an existing database
        :raises UnauthorizedToUseDatabase: If force is true this won't raise
        """
        try:
            requested_database = AccountDomain.get_requested_database()
        except NotADatabase as e:
            try:
                if e.body['requested_path'] not in current_app.config['RESOURCES_NOT_USING_DATABASES'] \
                        and not current_app.config['DOMAIN'][e.body['requested_path']]['use_default_database']:
                    raise e
            except KeyError:
                raise e
        else:
            if force or (requested_database in AccountDomain.actual['databases'] or AccountDomain.actual['role'] == Role.SUPERUSER):
                g.auth_requested_database = requested_database
                self.set_mongo_prefix(requested_database.replace("-", "").upper())
            else:
                raise UnauthorizedToUseDatabase()

    @staticmethod
    def get_requested_database_for_uri() -> str:
        try:
            return g.auth_requested_database + '/'
        except KeyError:
            return ''

    @staticmethod
    def set_needs_to_be_public(value: bool = True):
        g.needs_to_be_public = value

    @staticmethod
    def needs_to_be_public() -> bool:
        return getattr(g, 'needs_to_be_public', False)



"""
    Handles the submission of resources to external agents, like GRD.

    This package is divided in two:
    1. General purpose and extensible classes to handle submission.
    2. An implementation, grd_submitter, which handles submissions to GRD.

    The submission process is as follows:
    - Submitter_caller is instantiated, usually when instantiating DeviceHub, see flaskapp (app.grd_submitter_caller).
    Then:
    1. A hook starts the process by calling 'submit' of a SubmitterCaller instance (app.grd_submitter_caller).
    In grd_submitter, the hook is called after creating an event. The hooks passes the id of the resource.
    2. SubmitterCaller calls Submitter.
    3. Submitter gets the resource using the id, and uses an instance of Translator to transform the resource to fit
    the schema of the other agent. Finally performs log-in (see Auth class) and submits the transformed resource.

    Take a look at grd_submitter for an exemplifying implementation, extending the base Submitter, Translator and Auth
    classes.
"""


import pymongo
from ereuse_devicehub.resources.event.domain import EventDomain
from ereuse_devicehub.utils import Naming


class DeviceEventDomain(EventDomain):
    GENERIC_TYPES = {
        'Ready': {
            'fa': 'fa-check',
            'short_description': 'The devices work correctly, so they are ready to be used, sold, donated...',
            'sink': -2
        },
        'Repair': {
            'fa': 'fa-chain-broken',
            'short_description': 'The devices have been succesfully repaired',
            'sink': -1
        },
        'ToPrepare': {
            'fa': 'fa-wrench',
            'short_description': 'The devices need some maintenance, some kind of testing or preparation... to be ready',
            'sink': 1
        },
        'ToRepair': {
            'fa': 'fa-chain-broken',
            'short_description': 'The devices need repairing',
            'sink': 0
        },
        'ToDispose': {
            'fa': 'fa-trash-o',
            'short_description': 'The devices need to be taken for disposal.',
            'sink': -8
        },
        'Dispose': {
            'fa': 'fa-trash-o',
            'short_description': 'The devices have been successfully taken for disposal.',
            'sink': -9
        },
        'Free': {
            'fa': 'fa-shopping-cart',
            'short_description': 'The devices are available to be received for someone (sold, donated)',
            'sink': -4
        }
    }

    @staticmethod
    def add_prefix(type_name: str) -> str:
        """
        Adds a prefix to the type of a DeviceEvent (or subclass), if needed, and returns the full type with the prefix.

        It is used as coerce in @type field of DeviceEvent.
        """
        from ereuse_devicehub.resources.event.device.settings import DeviceEvent
        if type_name in (subclass.__name__ for subclass in DeviceEvent.subclasses()):
            return DeviceEventDomain.new_type(type_name)
        else:
            return str(type_name)

    @staticmethod
    def new_type(type_name):
        from ereuse_devicehub.resources.event.device.settings import DeviceEvent
        return Naming.new_type(type_name, DeviceEvent._settings['prefix'])

    @staticmethod
    def get_first_snapshot(device_id: str):
        return DeviceEventDomain.get_one(
            {'$query': {'device': device_id, '@type': 'devices:Snapshot'}, '$orderby': {'_created': pymongo.ASCENDING}})



import copy
import os
from pprint import pprint
from random import choice

from assertpy import assert_that
from bson import objectid
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.tests import TestStandard
from ereuse_devicehub.utils import Naming, coerce_type
from ereuse_devicehub.utils import NestedLookup


class TestSnapshot(TestStandard):
    DUMMY_DEVICES = (
        '1_1_Register_one_device_with_components',
        '1 - 2 - Register second device with components of first',
        '1 - 3 - Register 1-1 but removing motherboard moving processor from 1-2 to 1-1',
        '1 - 4 - Register 1-1 but removing processor (which ended again in 1-1) and adding graphicCard'
    )
    REAL_DEVICES = (
        'vostro', 'vaio', 'xps13'
    )
    RESOURCES_PATH = 'test_events/test_snapshot/resources/'

    def assertSimilarDevice(self, input_device: dict or str, created_device: dict or str):
        """
        Checks that the createdDevice is the same as the input one, removing computed values as hid... It uses etag.
        :param input_device Input device needs all the float values to have, by default, ".0", or it won't work
        """
        # todo make sure .0 doesn't crush in real program
        parsed_device = self.parse_device(input_device)
        with self.app.app_context():
            from ereuse_devicehub.resources.device.domain import DeviceDomain
            self.assertTrue(
                DeviceDomain.seem_equal(self.full(self.DEVICES, parsed_device),
                                        self.full(self.DEVICES, created_device)))

    def assertSimilarDevices(self, input_devices: list, created_devices: list, same_amount_of_devices=False):
        """
        Every created_device device must seem equal (same fields, except computed ones as hid...) as one of the input devices.

        There must be one input device per created device.
        :param input_devices:
        :param created_devices:
        :param same_amount_of_devices: bool Force to both lists to have the same amount of devices
        :return:
        """
        if same_amount_of_devices:
            self.assertEqual(len(input_devices), len(created_devices))
        for created_device in created_devices:
            found = False
            i = 0
            while not found and i < len(input_devices):
                try:
                    self.assertSimilarDevice(input_devices[i], created_device)
                    found = True
                except AssertionError:
                    pass
                i += 1
            self.assertTrue(found)

    def post_snapshot(self, input_snapshot):
        return self.post_and_check('{}/{}'.format(self.DEVICE_EVENT, self.SNAPSHOT), input_snapshot)

    def post_snapshot_get_full_events(self, input_snapshot, number_of_events_to_assert):
        snapshot = self.post_snapshot(copy.deepcopy(input_snapshot))
        self.assertEqual(len(snapshot['events']), number_of_events_to_assert)
        events = []
        for event_id in snapshot['events']:
            event, status_code = self.get('events', '', event_id)
            self.assert200(status_code)
            events.append(event)
        return events

    def creation(self, input_snapshot: dict, num_of_events: int = 1, do_second_time_snapshot=True) -> str:
        pprint("1st time snapshot:")
        events = self.post_snapshot_get_full_events(input_snapshot, num_of_events)
        self.assertLen(events, num_of_events)
        register = events[0]
        self.assertType(DeviceEventDomain.new_type('Register'), register)
        self.assertSimilarDevice(input_snapshot['device'], register['device'])
        if 'components' in input_snapshot:
            self.assertSimilarDevices(input_snapshot['components'], register['components'])
        # We do a snapshot again. We should receive a new snapshot without any event on it.
        if do_second_time_snapshot:
            pprint("2nd time snapshot:")
            snapshot, status_code = self.post('{}/{}'.format(self.DEVICE_EVENT, self.SNAPSHOT), input_snapshot)
            self.assert201(status_code)
            self.assertLen(snapshot['events'], num_of_events - 1)
        return register['device']

    def add_remove(self, input_snapshot):
        component = choice(input_snapshot['components'])
        found = False
        while not found:
            ignore_fields = self.app.config['DOMAIN'][Naming.resource(component['@type'])]['etag_ignore_fields']
            key = choice(list(component.keys()))
            found = key not in ignore_fields
        if type(component[key]) is int or type(component[key]) is float:
            component[key] += 10
        elif type(component[key]) is str:
            import uuid
            component[key] = uuid.uuid4().hex[:6].upper()
        events = self.post_snapshot_get_full_events(input_snapshot, 3)

    @staticmethod
    def get_num_events(snapshot):
        """
        Get the num of events a snapshot is going to produce, by knowing how many tests and erasures are in there.

        todo: compute add/Remove and other events, not just tests.
        :param snapshot:
        :return:
        """
        values = NestedLookup(snapshot, [], NestedLookup.key_equality_factory('test'))
        values += NestedLookup(snapshot, [], NestedLookup.key_equality_factory('erasure'))
        return len(values) + 1  # 1 == register event itself

    def test_add_remove(self):
        # todo create add/remove test with components and computers without hid
        # todo Check that the type of events generated are the correct ones
        folder = 'add_remove'
        snapshots = []
        for dummy_device in self.DUMMY_DEVICES:
            snapshots.append(self.get_fixture(folder, dummy_device))
        # We add the first device (2 times)
        self.creation(snapshots[0], self.get_num_events(snapshots[0]))
        # We register a new device, which has the processor of the first one
        # We have created 3 events (apart from the snapshot itself): Register, 1 Add and 1 Remove
        self.post_snapshot_get_full_events(snapshots[1], 3)
        # We register the first device again, but removing motherboard and moving processor from the second device
        # to the first. We have created 1 Add, 2 Remove (1. motherboard, 2. processor from second device)
        self.post_snapshot_get_full_events(snapshots[2], 3)
        # We register the first device but without the processor and adding a graphic card
        # We have created 1 Remove, 1 Add
        self.post_snapshot_get_full_events(snapshots[3], 2)

    def _test_snapshot_register_vostro(self):
        """
        Same as `test_snapshot_register_easy` however with real devices (fake serials), with all the risks that takes.
        :return:
        """
        self.creation(self.get_json_from_file(self.RESOURCES_PATH + 'vostro.json'), 2)

    def _test_snapshot_register_vaio(self):
        """
        Same as `test_snapshot_register_easy` however with real devices (fake serials), with all the risks that takes.
        :return:
        """
        self.creation(self.get_json_from_file(self.RESOURCES_PATH + self.REAL_DEVICES[1]))

    def _test_snapshot_register_dellxps(self):
        """
        Same as `test_snapshot_register_easy` however with real devices (fake serials), with all the risks that takes.
        :return:
        """
        self.creation(self.get_json_from_file(self.RESOURCES_PATH + self.REAL_DEVICES[2]))

    def test_snapshot_no_hid(self):
        """
        Tries a snapshot which's device has no hid, neither some components.

        The tests validates the process the process of inserting a device without hid.
        :return:
        """
        snapshot = self.get_fixture(self.SNAPSHOT, 'mounted')
        try:
            # Let's try first a simple snapshot
            self.post_snapshot(snapshot)
        except AssertionError as e:
            if e.args[0] == '422 != 201' and 'NeedsId' in e.message['_issues']['_id'][0]:
                # The system tells us that it could not register the device because the device (computer) has no hid
                # We can tell the system that this device already exists, by specifying an '_id', or stating
                # that this is new. We say it is new:
                snapshot['device']['forceCreation'] = True
                # And we repeat the process
                self.creation(snapshot, self.get_num_events(snapshot), False)
                # All ok. We remove the forceCreation and repeat the process
                del snapshot['device']['forceCreation']
                try:
                    self.post_snapshot(snapshot)
                except AssertionError as k:
                    # The system asks again the same. This time we will say that the device is the first one
                    # by specifying the '_id' to '1'
                    if k.args[0] == '422 != 201' and 'NeedsId' in k.message['_issues']['_id'][0]:
                        snapshot['device']['_id'] = '1'
                        # The system now is going to recognize the device and it's components,
                        # thus causing no extra event, apart from the snapshot itself
                        self.post_snapshot_get_full_events(snapshot, 0)
                    else:
                        raise e
            else:
                raise e
        else:
            self.assertTrue(False)  # We shouldn't we here, let's raise something

    def test_snapshot_real_devices(self) -> list:
        # todo the processor of mounted.json and xps13 generates the same hid, as S/N is 'To be filled...'
        for file_name in self.REAL_DEVICES:
            snapshot = self.get_fixture(self.SNAPSHOT, file_name)
            num_events = self.get_num_events(snapshot)
            self.creation(snapshot, num_events)

    def test_snapshot_2015_12_09(self, maximum: int = None, extra_fields_snapshot: dict = None):
        this_directory = os.path.dirname(os.path.realpath(__file__))
        file_directory = os.path.join(this_directory, 'resources', '2015-12-09')
        i = 0
        for filename in os.listdir(file_directory):
            if maximum is not None and i >= maximum:
                break
            if 'json' in filename:
                snapshot = self.get_json_from_file(filename, file_directory)
                snapshot.update(extra_fields_snapshot or {})
                num_events = self.get_num_events(snapshot)
                self.creation(snapshot, num_events)
                i += 1

    def test_benchmark(self):
        # todo add benchmark for processors (which is done in `test_erase_sectors`
        snapshot = self.get_fixture(self.SNAPSHOT, 'device_benchmark')
        device_id = self.creation(snapshot, self.get_num_events(snapshot))
        full_device, _ = self.get(self.DEVICES, '?embedded={"components": 1}', device_id)
        # Let's check that the benchmarks have been created correctly
        for component in full_device['components']:
            # benchmark is a writeonly value
            self.assertNotIn('benchmark', component)
            if component['@type'] in ('HardDrive', 'Processor'):
                self.assertIn('benchmarks', component)
                benchmark = next((c for c in snapshot['components'] if c['@type'] == component['@type']))['benchmark']
                # self.creation makes 2 post, so we will have 2 benchmarks that are exactly the same
                self.assertListEqual(component['benchmarks'], [benchmark, benchmark])

    def test_erase_sectors(self):
        """
        Tests EraseSectors.

        Inserts a device with BenchmarkProcessor and BenchmarkHardDrive.
        """
        self._test_erase_sectors('erase_sectors')

    def test_erase_sectors_many_steps(self):
        """
        The same as :func:`test_erase_sectors`, however with more sectors and different benchmarks.
        """
        self._test_erase_sectors('erase_sectors_steps')

    def _test_erase_sectors(self, fixture_name):
        snapshot = self.get_fixture(self.SNAPSHOT, fixture_name)
        # We do a Snapshot 2 times
        device_id = self.creation(snapshot, self.get_num_events(snapshot))
        full_device, _ = self.get(self.DEVICES, '?embedded={"components": 1}', device_id)
        found = False
        for component in full_device['components']:
            if 'erasures' in component:
                hard_drive, _ = self.get(self.DEVICES, '?embedded={"erasures":1}', component['_id'])
                # erasure is a writeonly value
                self.assertNotIn('erasure', hard_drive)
                # erasures must exist and contain an array with 2 erasures, which are the same
                coerce_type(snapshot)  # We add prefix to types in snapshot
                erasure = next((c for c in snapshot['components'] if
                                'serialNumber' in c and c['serialNumber'] == component['serialNumber']))['erasure']
                self.assertLen(hard_drive['erasures'], 2)
                assert_that(erasure).is_subset_of(hard_drive['erasures'][0])
                assert_that(erasure).is_subset_of(hard_drive['erasures'][1])
                found = True
        self.assertTrue(found, 'Any component has erasures!')

    def _test_signed(self):
        """
        The same as :func:`test_erase_sectors_many_steps`, but with a signed json.

        We just try to POST it one, we do not do a full stack of tests.
        """
        # todo get good public key
        raw_snapshot = self.get_fixture(self.SNAPSHOT, 'test_signed', False)
        self.post_snapshot(raw_snapshot)

    def test_703b6(self):
        """
        Tests a device generated with DeviceInventoroy version 7.0.3 beta 6
        """
        snapshot = self.get_fixture(self.SNAPSHOT, '703b6')
        num_events = self.get_num_events(snapshot)
        return self.creation(snapshot, num_events)

    def test_703b6_delete(self):
        """
        Tests deleting a full device, ensuring related events and components are deleted.
        """
        device_id = self.test_703b6()
        device, _ = self.get(self.DEVICES, '', device_id)
        # The materialization keeps events of the device and its components, we only get test_hd as an event
        # for a component, to try with it later; let's ignore the rest of events from components
        snapshot, test_hd, _, first_snapshot, _, _, register = (self.get(self.EVENTS, '', event['_id'])[0] for event in
                                                                device['events'])
        register, erase, test = (self.get(self.EVENTS, '', event_id)[0] for event_id in
                                 self.get(self.EVENTS, '', first_snapshot['_id'])[0]['events'])
        # Let's try deleting NOT the Snapshot that created the device
        self.delete_and_check(self.DEVICE_EVENT + '/snapshot/' + snapshot['_id'])
        _, status = self.get(self.EVENTS, '', snapshot['_id'])
        self.assert404(status)
        # We can still get the device, as the device was created with the *first* snapshot (that called Register)
        _, status = self.get(self.DEVICES, '', device_id)
        self.assert200(status)
        # The same for any other event that was not generated because of this snapshot
        _, status = self.get(self.EVENTS, '', test['_id'])
        self.assert200(status)
        # Or any component
        _, status = self.get(self.DEVICES, '', device['components'][0])
        self.assert200(status)

        # Let's add again a new snapshot
        # Note that this will produce a Remove event as there are no components
        new_snapshot = self.post_snapshot({'@type': 'Snapshot', 'device': device})
        new_remove, status = self.get(self.EVENTS, '', new_snapshot['events'][0])
        self.assert200(status)

        # Now let's delete the device
        self.delete_and_check(self.DEVICES + '/' + device_id)
        # The first Snapshot should have deleted all events that were generated from it
        _, status = self.get(self.EVENTS, '', test['_id'])
        self.assert404(status)
        _, status = self.get(self.EVENTS, '', erase['_id'])
        self.assert404(status)
        _, status = self.get(self.EVENTS, '', register['_id'])
        self.assert404(status)

        # Deleting the register event should have led to deleting the device,
        _, status = self.get(self.DEVICES, '', device_id)
        self.assert404(status)
        # ...all the components that were created in that Register,
        for component_id in device:
            _, status = self.get(self.DEVICES, '', component_id)
            self.assert404(status)
        # ...and finally the new snapshot we created...
        _, status = self.get(self.EVENTS, '', new_snapshot['_id'])
        self.assert404(status)
        # ...with the Remove it triggered...
        _, status = self.get(self.EVENTS, '', new_remove['_id'])
        self.assert404(status)
        # ...and the events of the components
        _, status = self.get(self.EVENTS, '', test_hd['_id'])
        self.assert404(status)

    def test_computer_monitor(self):
        snapshot = self.get_fixture(self.SNAPSHOT, 'monitor')
        self.creation(snapshot, 1)  # Register only

    def test_condition(self, maximum: int = None):
        condition = {
            'appearance': {'general': 'A'},
            'functionality': {'general': 'C'}
        }
        self.test_snapshot_2015_12_09(maximum or 5, {'condition': condition})

    def _test_giver(self, snapshot, account_email):
        device_id = self.creation(snapshot, 2, False)
        account, status = self.get(self.ACCOUNTS, '', account_email)
        self.assert200(status)
        # Let's see that the materialized snapshot contains the event
        device, _ = self.get(self.DEVICES, '', device_id)
        for event in device['events']:
            if event['@type'] == 'Snapshot':
                assert_that(event).contains_key('receiver')
                assert_that(event['from']).is_equal_to(account['_id'])
                assert_that(event['from']).is_type_of(objectid)

    def test_from_registered(self):
        """
            Tests that 'from' is set to the event.
        """
        snapshot = self.get_fixture(self.SNAPSHOT, 'vostro')
        account = {'email': 'r@r.com', 'name': 'R Registered', 'organization': 'R ORG', '@type': 'Account',
                   'databases': ['db1']}
        account = self.post_and_check(self.ACCOUNTS, account)
        snapshot['from'] = account['_id']
        self._test_giver(snapshot, 'r@r.com')

    def test_from_unregistered(self):
        """
            Tests that 'unregisteredFrom' is set to the event.
        """
        snapshot = self.get_fixture(self.SNAPSHOT, 'vostro')
        snapshot['from'] = {
            'email': 'r@r.com',
            'name': 'R Unregistered',
            'organization': 'R ORG'
        }
        self._test_giver(snapshot, 'r@r.com')

    def test_nice(self):
        snapshot = self.get_fixture(self.SNAPSHOT, 'nice')
        self.creation(snapshot, self.get_num_events(snapshot))

    def test_703b6_place_account(self):
        """
        Tests that the account and place in snapshot are correctly created.
        """
        snapshot = self.get_fixture(self.SNAPSHOT, '703b6')
        snapshot['from'] = {
            'email': 'hello@hello.com'
        }
        snapshot['place'] = self.post_and_check(self.PLACES, self.get_fixture(self.PLACES, 'place'))['_id']
        num_events = self.get_num_events(snapshot)
        device_id = self.creation(snapshot, num_events)
        device, _ = self.get(self.DEVICES, '', device_id)
        assert_that(device['place']).is_equal_to(snapshot['place'])
        account, status = self.get(self.ACCOUNTS, '', 'hello@hello.com')
        self.assert200(status)



from ereuse_devicehub.exceptions import SchemaError


class AlreadyAllocated(SchemaError):
    message = 'All the devices have already been allocated to this account, so nothing has been done.'
    field = 'to'
    pass



import copy

from ereuse_devicehub.resources.device.schema import Device
from ereuse_devicehub.resources.device.settings import DeviceSubSettings


class Mobile(Device):
    imei = {
        'type': 'string',
        'unique': True
    }
    meid = {
        'type': 'string',
        'unique': True
    }
    type = {
        'type': 'string',
        'allowed': {'Smartphone', 'Tablet'},
        'required': True
    }
    manufacturer = copy.copy(Device.manufacturer)
    manufacturer['required'] = True
    serialNumber = copy.copy(Device.serialNumber)
    serialNumber['required'] = True
    model = copy.copy(Device.model)
    model['required'] = True


class MobileSettings(DeviceSubSettings):
    _schema = Mobile



from ereuse_devicehub.resources.event.device.settings import components, EventWithOneDevice, EventSubSettingsOneDevice


class Add(EventWithOneDevice):
    components = components


class AddSettings(EventSubSettingsOneDevice):
    _schema = Add
    glyphicon = 'fa-plus-square-o'
    short_description = 'Components have been added to a device'
    item_methods = ['GET']



import copy

import iso3166
import pymongo

from ereuse_devicehub.resources.resource import ResourceSettings
from ereuse_devicehub.resources.schema import Thing


class Place(Thing):
    geo = {
        'type': 'polygon',
        'sink': -5,
        'description': 'Set the area of the place. Be careful! Once set, you cannot update the area.',
        'modifiable': False
    }
    type = {
        'type': 'string',
        'allowed': {'Department', 'Zone', 'Warehouse', 'CollectionPoint'}
    }
    devices = {
        'type': 'list',
        'schema': {
            'type': 'string',
            'data_relation': {
                'resource': 'devices',
                'field': '_id',
                'embeddable': True
            }
        },
        'default': [],
        'unique_values': True
    }
    byUser = {
        'type': 'objectid',
        'data_relation': {
            'resource': 'accounts',
            'field': '_id',
            'embeddable': True
        },
        'readonly': True
    }
    address = {
        'type': 'dict',
        'schema': {
            'addressCountry': {
                'type': 'string',
                'allowed': {key for key in iso3166.countries_by_alpha2},
                'description': 'The name of the country',
                'doc': 'The addressCountry as per ISO 3166 (2 characters).'
            },
            'addressLocality': {
                'type': 'string',
                'description': 'The locality. For example, Barcelona.'
            },
            'addressRegion': {
                'type': 'string',
                'description': 'The region. For example, CA.'
            },
            'postalCode': {
                'type': 'string',
                'description': 'The postal code. For example, 94043.'
            },
            'streetAddress': {
                'type': 'string',
                'description': 'The street address. For example, C/Jordi Girona, 1-3.'
            }
        }
    }
    telephone = {
        'type': 'string'
    }

    label = copy.deepcopy(Thing.label)


Place.label['required'] = True


class PlaceSettings(ResourceSettings):
    resource_methods = ['GET', 'POST']
    item_methods = ['GET', 'PATCH', 'DELETE', 'PUT']
    _schema = Place
    datasource = {
        'default_sort': [('_created', -1)],
        'source': 'places'
    }
    extra_response_fields = ResourceSettings.extra_response_fields + ['devices']
    mongo_indexes = {
        'geo': [('components', pymongo.GEO2D)]
    }



from assertpy import assert_that

from ereuse_devicehub.tests import TestBase
from ereuse_devicehub.utils import Naming


class TestResourceNaming(TestBase):
    @staticmethod
    def try_prefix(original_type, prefix, supposed_resource_name):
        """
        :param original_type: The resource type
        :param prefix:
        :param supposed_resource_name: What should the resource name be?
        """
        type_name = Naming.new_type(original_type, prefix)
        equal = original_type if prefix is None else '{}{}{}'.format(prefix, Naming.TYPE_PREFIX, original_type)
        assert_that(type_name).is_equal_to(equal)
        resource_name = Naming.resource(type_name)
        equal = supposed_resource_name if prefix is None else '{}{}{}'.format(prefix, Naming.RESOURCE_PREFIX,
                                                                              supposed_resource_name)
        assert_that(resource_name).is_equal_to(equal)

    def test_prefix(self):
        """
            Tests the creation of prefixes, given a type.
        """
        # Let's try first with a type that doesn't change the number (singular - plural)
        self.try_prefix('Snapshot', 'devices', 'snapshot')
        # Now with a type that can change its number (see RESOURCES_CHANGING_NUMBER in settings)
        self.try_prefix('Event', 'projects', 'events')
        # Finally without prefix
        self.try_prefix('Device', None, 'devices')



from multiprocessing import Process, Queue

from eve.methods.post import post_internal
from flask import json
from pymongo.errors import DuplicateKeyError

from ereuse_devicehub.resources.account.role import Role
from ereuse_devicehub.resources.submitter.submitter import Submitter
from ereuse_devicehub.utils import get_last_exception_info


class SubmitterCaller:
    """
    Prepares a submitter in a long-running daemon process, sending events through 'submit'.
    """
    token = None
    """
        The token of the account submitters use to access DeviceHub. It is shared to avoid
        performing login many times.
    """

    def __init__(self, app: 'DeviceHub', submitter: Submitter):
        """
        :param submitter: The submitter class to invoke when called
        """
        self.submitter = submitter
        self.queue = Queue()
        self.process = None
        self.app = app
        if not self.token:  # Maybe it has been already set by another submitter (class attribute)
            self.token = self.prepare_user(app)
        self.prepare_process()

    def submit(self, event_id: str, database: str, resource_name: str, *args):
        """
        Enqueues the event to be submitted through the submitter, in another thread.
        :param event_id: The id of the event to submit. The event is searched in the database through GET.
        :param database: The database to get the event from.
        :param resource_name: The type of the resource in 'resource_name'format.
        :param args: An optional and extra list of optional args to supply to the submitter.
        """
        self.prepare_process()  # We ensure the process is there, and we start it if it died
        self.queue.put((event_id, database, resource_name) + args)

    def prepare_process(self):
        """
        Ensures that the process is up and ready, or prepares one otherwise.
        """
        if not self.process or not self.process.is_alive():
            # noinspection PyArgumentList
            self.process = Process(target=_process,
                                   args=(self.queue, self.token, self.app, self.submitter),
                                   daemon=True)
            self.process.start()

    @staticmethod
    def prepare_user(app):
        with app.test_request_context():
            account = app.config['SUBMITTER_ACCOUNT']
            account.update({'role': Role.SUPERUSER})
            account['@type'] = 'Account'
            account['databases'] = app.config['DATABASES']
            if app.data.find_one_raw('accounts', {'email': account['email']}) is None:
                try:
                    post_internal('accounts', dict(account), True)  # If we validate, RolesAuth._set_database changes our db
                except DuplicateKeyError:
                    pass
        response = app.test_client().post('login', data=json.dumps(account), content_type='application/json')
        js = json.loads(response.data.decode())
        return js['token']


def _process(queue: Queue, token: str, app, submitter_class=Submitter):
    """
        A representation of a separate process.
        It's a for_all: It blocks waiting for events to log, and when there is an event, it invokes the submitter.
    :param queue: The queue sent from SubmitterCaller.
    :param token: The token of the DeviceHub user Submitter is going to use to get data.
    :param submitter_class: The Submitter *class* to instantiate and use.
    """
    submitter = submitter_class(token, app)
    while True:
        try:
            submitter.submit(*queue.get(True))  # We block ourselves waiting for something in the queue
        except Exception as e:
            submitter.logger.error(get_last_exception_info())
            raise e



from eve.auth import requires_auth
from flask import jsonify
from flask import request

from ereuse_devicehub.aggregation.aggregation import Aggregation, AggregationError
from ereuse_devicehub.flask_decorators import crossdomain, cache


@cache(Aggregation.CACHE_TIMEOUT)
@crossdomain(origin='*', headers=['Content-Type', 'Authorization'])
@requires_auth('resource')
def aggregate_view(db, resource, method):
    """
    Performs a login. We make this out of eve, being totally open.
    :return:
    """
    aggregation = Aggregation(resource)
    if method == '_aggregate':
        raise AggregationError("You cannot use _aggregate.")
    try:
        m = getattr(aggregation, method)(request.args)
    except AttributeError as a:
        raise AggregationError(a.args)
    else:
        return jsonify(m)



import random
import string

from ereuse_devicehub.resources.account.domain import AccountDomain, UserNotFound
from ereuse_devicehub.resources.account.role import Role
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.rest import execute_post
from flask import current_app as app
from passlib.handlers.sha2_crypt import sha256_crypt


def hash_password(accounts: list):
    for account in accounts:
        if 'password' in account:
            account['password'] = sha256_crypt.encrypt(account['password'])


def add_token(documents: list):
    for document in documents:
        token = generate_token()
        while app.data.find_one_raw('accounts', {'token': token}) is not None:
            token = generate_token()
        document["token"] = token


def generate_token() -> str:
    return (''.join(random.choice(string.ascii_uppercase)
                    for x in range(10)))


# noinspection PyPep8Naming
def set_byUser(resource_name: str, items: list):
    if 'byUser' in app.config['DOMAIN'][resource_name]['schema']:
        for item in items:
            if 'byUser' not in item:
                item['byUser'] = AccountDomain.actual['_id']


# noinspection PyPep8Naming
def set_byOrganization(resource_name: str, items: list):
    """
    Sets the 'byOrganization' field, which is the materialization of the organization of byUser (the actual user).
    This materialization shall no be updated when the user's organization changes.
    """
    for item in items:
        if 'byOrganization' in app.config['DOMAIN'][resource_name]['schema']:
            if 'organization' in AccountDomain.actual:
                item['byOrganization'] = AccountDomain.actual['organization']


def set_default_database_if_empty(accounts: list):
    for account in accounts:
        if 'defaultDatabase' not in account and account['role'] != Role.SUPERUSER:
            account['defaultDatabase'] = account['databases'][0]


def add_or_get_inactive_account_receive(events: list):
    for event in events:
        _add_or_get_inactive_account_id(event, 'receiver')


def add_or_get_inactive_account_allocate(events: list):
    for event in events:
        _add_or_get_inactive_account_id(event, 'to')


def add_or_get_inactive_account_snapshot(events: list):
    for event in events:
        _add_or_get_inactive_account_id(event, 'from')


def _add_or_get_inactive_account_id(event, field_name):
    """
    We need to execute after insert and insert_resource.
    """
    if field_name in event and type(event[field_name]) is dict:
        try:
            # We look for just accounts that share our database
            _id = AccountDomain.get_one({
                'email': event[field_name]['email'] # 'databases': {'$in': AccountDomain.actual['databases']} todo review this
            })['_id']
        except UserNotFound:
            event[field_name]['databases'] = [AccountDomain.get_requested_database()]
            event[field_name]['active'] = False
            event[field_name]['@type'] = 'Account'
            _id = execute_post('accounts', event[field_name], True)['_id']
        event[field_name] = _id



"""
This script is thought to run in a folder with JSON with the objective of:
- Add the json files to the test suite
- Make JSON devices anonymous by replacing S/N with hash values. Hash values are injective methods.
- Detect special characters in HID fields (S/N, model, manufacturer) which would cause a problem
"""
import hashlib
import json
import os
from collections import defaultdict, OrderedDict
from pprint import pprint

HID_FIELDS = 'serialNumber', 'manufacturer', 'model'
chars = defaultdict(OrderedDict)
words = defaultdict(list)
this_directory = os.path.dirname(os.path.realpath(__file__))


def get_json_from_file(filename: str) -> dict:
    with open(os.path.abspath(os.path.join(this_directory, filename))) as data_file:
        return json.load(data_file)


def sanitize(value: str, key: str) -> str:
    for char in value:
        chars[key][char] = chars[key][char] + 1 if char in chars[key] else 0
    words[key].append(value)
    if key == 'serialNumber':
        return hashlib.sha224(value.encode()).hexdigest()
    else:
        return value


def sanitize_event(d: dict):
    """
    Modified from http://stackoverflow.com/a/11700817
    :param d:
    :return:
    """
    new = {}
    for k, v in d.items():
        if isinstance(v, dict):
            v = sanitize_event(v)
        elif k == 'components':
            new_v = []
            for lv in v:
                new_v.append(sanitize_event(lv))
            v = new_v
        elif k in HID_FIELDS:
            v = sanitize(v, k)
        new[k] = v
    return new


def main():
    dest_dict = os.path.join(this_directory, 'new')
    os.makedirs(dest_dict, exist_ok=True)
    for filename in os.listdir('.'):
        if 'json' in filename:
            event = get_json_from_file(filename)
            del event['debug']
            new_event = sanitize_event(event)
            with open(os.path.join(dest_dict, new_event['label'] + '.json'), 'xt') as f:
                f.write(json.dumps(new_event))
    with open(os.path.join(dest_dict, 'chars.txt'), 'xt') as f:
        pprint(chars, f)
    with open(os.path.join(dest_dict, 'words.txt'), 'xt') as f:
        pprint(words, f)


main()



from ereuse_devicehub.validation.validation import DeviceHubValidator

"""Grades the state of the device in different areas."""
condition = {
    'appearance': {
        'type': 'dict',
        'schema': {
            'general': {
                'type': 'string',
                'allowed': DeviceHubValidator.SCALE_AD,
                'description': 'Grades the imperfections that aesthetically affect the device, but not its usage.'
            }
        }
    },
    'functionality': {
        'type': 'dict',
        'schema': {
            'general': {
                'type': 'string',
                'allowed': DeviceHubValidator.SCALE_AD,
                'description': 'Grades the defects of a device that affect its usage.'
            }
        }
    }
}



from flask import current_app

from ereuse_devicehub.resources.event.device.settings import Event
from ereuse_devicehub.security.authentication import AccountDomain


def submit_events_to_grd(resource: str, events: list):
    """
    Send info from the hook to the Logger in its thread
    """
    if current_app.config.get('GRD', True) and \
                    resource in current_app.config['EVENTS_IN_GRD'] and resource in Event.resource_types:
        for event in events:
            current_app.grd_submitter_caller.submit(str(event['_id']), AccountDomain.get_requested_database(), resource)



from assertpy import assert_that

from ereuse_devicehub.tests import TestBase
from ereuse_devicehub.utils import Naming


class TestDeviceEventBasic(TestBase):
    """
        Tests that take care of the creation and configuration of device events.
    """
    def test_creation(self):
        """
        Tests that device events have been created correctly, taking special care of:
        - All events have been created
        - @type and type
        - URL
        - prefix
        """
        events = ('add', 'allocate', 'deallocate', 'dispose', 'free',
                  'locate', 'ready', 'receive', 'register', 'remove', 'repair', 'snapshot', 'test-hard-drive',
                  'to-dispose', 'to-repair', 'to-prepare')  # all subclasses from DeviceEvent in resource type
        events = ['{}{}{}'.format('devices', Naming.RESOURCE_PREFIX, event) for event in events]  # we prefix them
        events += ['accounts', 'devices', 'computer']  # We check some non-prefixed regular resources...
        assert_that(self.domain).contains_key(*events)
        # Type of snapshot should be 'devices:Snapshot'
        snapshot = self.domain['{}{}{}'.format('devices', Naming.RESOURCE_PREFIX, 'snapshot')]
        assert_that(snapshot['schema']['@type']['allowed'])\
            .is_equal_to({'{}{}{}'.format('devices', Naming.TYPE_PREFIX, 'Snapshot')})
        devices = self.domain['devices']
        # And any other type not subclass from DeviceEvent should be without prefix
        assert_that(devices['schema']['@type']['allowed']).contains('Device', 'Computer', 'HardDrive')  # ...and more

        # Checking that the url generated contains 'devices' for DeviceEvent...
        assert_that(snapshot['url']).is_equal_to('events/devices/snapshot')
        # ...but it doesn't add devices to others (it would be then 'devices/devices')
        assert_that(devices['url']).is_equal_to('devices')



from ereuse_devicehub.resources.device.domain import DeviceDomain
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.resources.event.domain import EventDomain


class ComponentDomain(DeviceDomain):
    @classmethod
    def get_parent(cls, _id: str) -> dict or None:
        return cls.get_one({'components': {'$in': [_id]}})

    @classmethod
    def get_similar_component(cls, component: dict, parent_id: str) -> dict:
        """Gets a component that has same parent, doesn't generate HID and their ETAG are the same"""
        # We the unsecured _id of the devices of all parent_id snapshots
        snapshots = EventDomain.get({'@type': DeviceEventDomain.new_type('Snapshot'), 'device': parent_id})
        devices_id = set()
        for snapshot in snapshots:
            for unsecured in snapshot['unsecured']:
                devices_id.add(unsecured['_id'])
        # We get the devices whose _id and etag matches
        etag = cls.generate_etag(component)
        query = {'_id': {'$in': list(devices_id)}, '_etag': etag}
        device = cls.get_one(query)
        return cls.get_one(device['_id'])  # todo if we materialize components we do not need to do double query

    @classmethod
    def get_devices_with_components(cls, devices_id: list) -> list:
        """
        Gets a list of devices with their components, not more values.
        :param devices_id:
        :return:
        """
        return cls.get({'_id': {'$in': devices_id}})

    @classmethod
    def get_components_in_set(cls, devices_id: list) -> set:
        """
        For a given list of devices, gets one set containing the union of all of their components
        :param devices_id:
        :return:
        """
        components = set()
        for device in cls.get_devices_with_components(list(devices_id)):
            if 'components' in device:
                components |= set(device['components'])
        return components

    @classmethod
    def benchmark(cls, component: dict):
        """
        Adds the result of a benchmark to a component.

        Components needs to exist in the database.
        :param component:
        """
        assert '_id' in component, 'Component needs to be created'
        if 'benchmark' in component:
            cls.update_raw(component['_id'], {'$push': {'benchmarks': component['benchmark']}})



import copy

from ereuse_devicehub.resources.event.device.settings import place, components, EventWithDevices, \
    EventSubSettingsMultipleDevices


class Locate(EventWithDevices):
    place = place
    components = copy.deepcopy(components)


Locate.components['readonly'] = True
Locate.geo = copy.deepcopy(Locate.geo)
Locate.geo['excludes'] = 'place'  # geo xor place
Locate.geo['or'] = ['place']


class LocateSettings(EventSubSettingsMultipleDevices):
    _schema = Locate
    fa = 'fa-map-marker'
    sink = -3
    short_description = 'The devices have been placed.'



from ereuse_devicehub.resources.event.device.domain import DeviceEventDomain
from ereuse_devicehub.resources.event.domain import EventNotFound
from ereuse_devicehub.resources.place.domain import PlaceDomain, CannotDeleteIfHasEvent


def set_place_in_devices(items: list):
    for item in items:
        if 'devices' in item:
            PlaceDomain.update_devices(set(), set(item['devices']), item['_id'])


def update_place_in_devices_if_places(updated_place: dict, original_place: dict):
    if 'devices' in updated_place:  # For patch, if no value, it means it is not being updated
        update_place_in_devices(updated_place, original_place)


def update_place_in_devices(replaced_place: dict, original_place: dict):
    PlaceDomain.update_devices(set(original_place['devices']), set(replaced_place['devices']), replaced_place['_id'])


def unset_place_in_devices(place):
    PlaceDomain.update_devices(set(place['devices']), set(), None)


def avoid_deleting_if_has_event(item):
    try:
        DeviceEventDomain.get_one({'place': item['_id']})
    except EventNotFound:
        pass
    else:
        raise CannotDeleteIfHasEvent()



from setuptools import setup, find_packages

setup(
    name='eReuse-DeviceHub',
    version='0.1.1',
    packages=find_packages(exclude=('contrib', 'docs', 'scripts')),
    url='https://github.com/eReuse/DeviceHub',
    license='AGPLv3 License',
    author='eReuse.org team',
    author_email='x.bustamante@ereuse.org',
    description='DeviceHub is a system to manage devices focused in reusing them. ',
    install_requires=[
        'inflection>=0.3.1,<0.4',
        'eve>=0.6.3,<7.0',
        'passlib>=1.6.5,<2.0',
        'validators>=0.10,<0.20',
        'requests>=2.9.1,<3.0',
        'python-gnupg>=0.3.8,<0.4',  # To use gnupg, install gpg2
        'flask-cache>=0.13.1',
        'python-gnupg',
        'iso3166'
    ],
    keywords='eReuse.org DeviceHub devices devicehub reuse recycle it asset management',
    # http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-extras-optional-features-with-their-own-dependencies
    extras_require={
        'docs': [
            'sphinx>=1.4.6,<1.4.7',
            'sphinxcontrib-httpdomain>=1.5'
        ]
    },
    # http://setuptools.readthedocs.io/en/latest/setuptools.html#test-build-package-and-run-a-unittest-suite
    test_suite='ereuse_devicehub.tests',
    test_require=[
        'assertpy'
    ],
    include_package_data=True,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Manufacturing',
        'License :: OSI Approved :: GNU Affero General Public License v3',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Office/Business',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content'
    ]
)



from ereuse_devicehub.resources.device.component.settings import Component, ComponentSubSettings
from ereuse_devicehub.resources.schema import UnitCodes


class RamModule(Component):
    size = {
        'type': 'integer',
        'unitCode': UnitCodes.mbyte,
        'min': 1,
        'sink': 1
    }
    speed = {
        'type': 'float',
        'unitCode': UnitCodes.mhz,
        'min': 1,
        'sink': -1
    }


class RamModuleSettings(ComponentSubSettings):
    _schema = RamModule



from bson import json_util
from ereuse_devicehub.exceptions import InnerRequestError
from ereuse_devicehub.resources.device.component.domain import ComponentDomain
from ereuse_devicehub.resources.device.domain import DeviceDomain
from ereuse_devicehub.resources.device.exceptions import DeviceNotFound, NoDevicesToProcess
from ereuse_devicehub.rest import execute_post, execute_delete
from ereuse_devicehub.utils import Naming
from eve.methods.delete import deleteitem_internal
from flask import current_app as app


def post_devices(registers: list):
    """
    Main function for Register. For the given devices, POST the new ones.

    If there is an unexpected error (device has a bad field, for example), it undoes all posted devices. In db terms,
        the commit is all the registers.

    If the function is called by post_internal(), as the method keeps the reference of the passed in devices, the
        caller will see how their devices are replaced by the db versions, plus a 'new' property acting as a flag
        to indicate if the device is new or not.

    :type registers: list
    :raises InnerRequestError: for any error provoked by a failure in the POST of a device (except if the device already
        existed). It carries the original error sent by the POST.
    """
    log = []
    for register in registers:
        caller_device = register['device']  # Keep the reference from where register['device'] points to
        _execute_register(caller_device, log)
        register['device'] = caller_device['_id']  # Change the reference of register['device'], but not caller_device
        if 'components' in register:
            caller_components = register['components']
            register['components'] = []
            for component in caller_components:
                component['parent'] = caller_device['_id']
                _execute_register(component, log)
                if 'new' in component:  # todo put new in g., don't use device
                    register['components'].append(component['_id'])
            if not register['components'] and 'new' not in caller_device:
                _abort(
                    log)  # If we have not POST neither any component and any device there is no reason for register to exist
            if 'new' in caller_device:
                set_components(register)
    from ereuse_devicehub.resources.hooks import set_date
    set_date(None, registers)  # Let's get the time AFTER creating the devices


# noinspection PyUnboundLocalVariable
def _execute_register(device: dict, log: list, force_new=False):
    """

    :param device:
    :param log:
    :param force_new: Raises exception if the device is not new. Debugging parameter.
    """
    device['hid'] = 'dummy'
    try:
        db_device = execute_post(Naming.resource(device['@type']), device)
    except InnerRequestError as e:
        if force_new:
            raise e
        new = False
        try:
            db_device = _get_existing_device(e)
            # We add a benchmark todo move to another place?
            device['_id'] = db_device['_id']
            ComponentDomain.benchmark(device)
        except DeviceNotFound:
            _abort(log, e)
    else:
        log.append(db_device)
        new = True
    device.clear()
    device.update(
        db_device)  # We do not assign so we preserve the reference todo db_device just have extra_post_fields, not all fields of device
    if new:
        device['new'] = new


def _get_existing_device(e):
    device = None
    for field in 'hid', '_id', 'model':  # unique fields
        if field in e.body['_issues']:
            try:
                for error in e.body['_issues'][field]:
                    try:
                        device = json_util.loads(error)['NotUnique']
                    except (ValueError, KeyError):
                        pass
                    else:
                        break
            except (ValueError, KeyError):  # it can be an unique field but the
                raise DeviceNotFound()
    if not device:
        raise DeviceNotFound()
    return device


def set_components(register):
    """
    Sets the new devices to the materialized attribute 'components' of the parent device.
    """
    app.data.driver.db['devices'].update(
        {'_id': register['device']},
        {'$set': {'components': register['components']}}
    )


def _abort(log, e: Exception = NoDevicesToProcess()):
    for device in reversed(log):
        deleteitem_internal(Naming.resource(device['@type']), device)
    raise e


def delete_device(resource_name: str, register):
    if register.get('@type') == 'devices:Register':
        for device_id in [register['device']] + register.get('components', []):
            execute_delete(Naming.resource(DeviceDomain.get_one(device_id)['@type']), device_id)



class Role:
    """
    Roles specify what users can do generally.
    Locations alter what users can do. Locations permissions > roles. Admins don't get affected by location perms.
    Roles are ordered hierarchically. This means that an amateur role has the same permissions as a basic, and more.
    An employee can do the same as an amateur, and more.
    The gradient is:
    BASIC < AMATEUR < EMPLOYEE < ADMIN < SUPERUSER

    We can use operators to compare between roles. For example BASIC < AMATEUR == True
    """
    BASIC = 'basic'
    """Most basic user. Cannot do anything (except its account). Useful for external people."""
    AMATEUR = 'amateur'
    """Can create devices, however can just see and edit the ones it created. Events are restricted."""
    EMPLOYEE = 'employee'
    """Technicians. Full spectre of operations. Can interact with devices of others."""
    ADMIN = 'admin'
    """worker role + manage other users (except superusers). No location restrictions. Can see analytics."""
    SUPERUSER = 'superuser'
    """admin + they don't appear as public users, and they can manage other superusers. See all databases."""
    ROLES = BASIC, AMATEUR, EMPLOYEE, ADMIN, SUPERUSER
    """In grading order (BASIC < AMATEUR)"""
    MANAGERS = ADMIN, SUPERUSER

    def __init__(self, representation):
        if representation not in self.ROLES:
            raise TypeError(representation + ' is not a role.')
        self.role = representation

    def is_manager(self):
        return self.role in self.MANAGERS

    def has_role(self, roles: set):
        return self.role in roles

    def __lt__(self, other):
        if isinstance(other, Role):
            return self.ROLES.index(self.role) < self.ROLES.index(other.role)
        elif isinstance(other, str):
            return self.ROLES.index(self.role) < self.ROLES.index(other)
        else:
            raise NotImplemented

    def __le__(self, other):
        if isinstance(other, Role):
            return self.ROLES.index(self.role) <= self.ROLES.index(other.role)
        elif isinstance(other, str):
            return self.ROLES.index(self.role) <= self.ROLES.index(other)
        else:
            raise NotImplemented

    def __gt__(self, other):
        return not self.__le__(other)

    def __ge__(self, other):
        return not self.__lt__(other)

    def __eq__(self, other):
        if isinstance(other, Role):
            return self.role == other.role
        elif isinstance(other, str):
            return self.role == other
        else:
            raise NotImplemented

    def __hash__(self):
        return self.role.__hash__()


import base64

from bson.objectid import ObjectId
from flask import current_app
from flask import g
from flask import request
from werkzeug.http import parse_authorization_header

from ereuse_devicehub.exceptions import WrongCredentials, BasicError, StandardError
from ereuse_devicehub.resources.account.role import Role
from ereuse_devicehub.resources.account.settings import AccountSettings
from ereuse_devicehub.resources.domain import Domain, ResourceNotFound
from ereuse_devicehub.utils import ClassProperty


class AccountDomain(Domain):
    resource_settings = AccountSettings
    @staticmethod
    def get_requested_database():
        requested_database = request.path.split('/')[1]
        if requested_database not in current_app.config['DATABASES']:
            raise NotADatabase({'requested_path': requested_database})
        else:
            return requested_database

    # noinspection PyNestedDecorators
    @ClassProperty
    @classmethod
    def actual(cls) -> dict:
        try:
            if not hasattr(g, '_actual_user'):
                from flask import request
                try:
                    x = request.headers.environ['HTTP_AUTHORIZATION']
                    token = parse_authorization_header(x)['username']
                    from flask import current_app as app
                    g._actual_user = AccountDomain.get_one({'token': token})
                    g._actual_user['role'] = Role(g._actual_user['role'])
                except UserNotFound:
                    raise UserIsAnonymous("You need to be logged in.")
                except TypeError:
                    raise NoUserForGivenToken()
            return g._actual_user
        except RuntimeError as e:
            # Documentation access this variable
            if str(e) != 'working outside of application context':
                raise e

    @classmethod
    def get_one(cls, id_or_filter: dict or ObjectId or str):
        try:
            return super().get_one(id_or_filter)
        except ResourceNotFound:
            raise UserNotFound()

    @staticmethod
    def import_key(key: str) -> str:
        """
        Imports the key for the user
        :param key: GPG Public Key
        :raises CannotImportKey:
        :return: Fingerprint of the imported key
        """
        result = current_app.gpg.import_keys(key)
        if result.count == 0:
            raise CannotImportKey()
        return result.fingerprint[0]

    @staticmethod
    def hash_token(token):
        return base64.b64encode(
            str.encode(token + ':'))  # Framework needs ':' at the end before send it to client


class UserIsAnonymous(WrongCredentials):
    pass


class NoUserForGivenToken(WrongCredentials):
    pass


class NotADatabase(BasicError):
    status_code = 400


class CannotImportKey(StandardError):
    status_code = 400
    message = "We could not import the key. Make sure it is a valid GPG Public key."


class UserNotFound(ResourceNotFound):
    pass


from ereuse_devicehub.resources.device.component.settings import Component, ComponentSubSettings


class SoundCard(Component):
    pass


class SoundCardSettings(ComponentSubSettings):
    _schema = SoundCard



from unittest.mock import MagicMock

from assertpy import assert_that

from ereuse_devicehub.tests import TestStandard


class TestSubmitter(TestStandard):
    """
        Tests submitter through the GRD implementation.
    """

    def test_creation(self):
        assert_that(hasattr(self.app, 'grd_submitter_caller')).is_true()
        assert_that(hasattr(self.app, 'submit_events_to_grd')).is_true()
        assert_that(self.app.grd_submitter_caller.token).is_not_none()

    @staticmethod
    def mock_submit(submitter_class, token, app, mock_post):
        """
            Mocks Submitter._post and SubmitterCaller.submit, asserting the values that would have been sent to GRD.
        :return:
        """

        submitter = submitter_class(token, app)
        submitter._post = MagicMock()
        submitter._post.side_effect = mock_post

        def _mock_submit(event_id: str, database: str, resource_name: str, *args):
            """Mocks SubmitterCaller.submit, directly calling Submitter."""
            submitter.submit(event_id, database, resource_name, *args)
            assert_that(submitter._post.call_count).is_equal_to(1)

        return _mock_submit

    @staticmethod
    def mock_post(test, grd_resource, grd_url):
        def _mock_post(resource: dict, url: str):
            """Mocks Submitter._post, asserting entering values"""
            assert_that(resource['url']).contains('https://www.example.com/dht1/events/')
            assert_that(resource['byUser']).contains('https://www.example.com/accounts/')
            del resource['url']
            del resource['byUser']
            del resource['dhDate']
            assert_that(grd_resource).is_subset_of(resource)
            assert_that(url).is_equal_to(grd_url)
        return _mock_post

    def test_snapshot(self):
        # We load the expected result that is sent to GRD
        grd_resource = self.get_fixture(self.SNAPSHOT, 'vaio_grd')
        grd_url = 'https://sandbox.ereuse.org/api/devices/register'  # todo ensure requeriment
        # We prepare the call by mocking through mock_submit
        submitter_caller = self.app.grd_submitter_caller
        submitter_caller.submit = MagicMock()
        submitter_caller.submit.side_effect = self.mock_submit(submitter_caller.submitter, submitter_caller.token,
                                                               submitter_caller.app,
                                                               self.mock_post(self, grd_resource, grd_url))
        # We create the device, generating the devices:Register submission
        self.post_fixture(self.SNAPSHOT, '{}/{}'.format(self.DEVICE_EVENT, self.SNAPSHOT), 'vaio')
        assert_that(submitter_caller.submit.call_count).is_equal_to(1)



import copy

from ereuse_devicehub.resources.event.device.settings import components, EventWithDevices, \
    EventSubSettingsMultipleDevices


class Deallocate(EventWithDevices):
    _from = {
        'type': 'objectid',
        'data_relation': {
            'resource': 'accounts',
            'field': '_id',
            'embeddable': True,
        },
        'sink': 2
    }
    fromOrganization = {
        # Materialization of the organization that, by the time of the deallocation, the user worked in
        'type': 'string',
        'readonly': True
    }

    @classmethod
    def _clean(cls, attributes: dict, attributes_to_remove: tuple = None) -> dict:
        full_dict = super(Deallocate, cls)._clean(attributes, attributes_to_remove)
        full_dict['from'] = full_dict['_from']
        del full_dict['_from']
        return full_dict


Deallocate.components = copy.deepcopy(components)
Deallocate.components['readonly'] = True


# Receiver OR ReceiverEmail. We need to hook this in a required field so it is always executed
# And @type is an always required field so we can happily hook on it

class DeallocateSettings(EventSubSettingsMultipleDevices):
    _schema = Deallocate
    fa = 'fa-hand-o-left'
    sink = -6
    extra_response_fields = EventSubSettingsMultipleDevices.extra_response_fields + ['from']
    short_description = 'The opposite of Allocate. Remove the ownership of the devices from someone'



import validators
from bson import ObjectId
from flask import current_app

from ereuse_devicehub.exceptions import SchemaError


class Coercer:
    """Contains general 'coerces' used by validation"""

    @staticmethod
    def url_to_id(url, field_name, document, schema):
        """
        Given an URL, looks for a resource with the same URL in the sameAs field. If succeeds, returns the internal id,
        otherwise throws error. If the value is not an URL it only returns the value.
        :param url:
        :param field_name:
        :param document:
        :return: the internal (DeviceHub) identifier of the resource. Note that ObjectId are parsed to str to mimic JSON.
        """
        if validators.url(str(url)):
            data_relation = schema[field_name]['data_relation']
            response = current_app.data.find_one_raw(data_relation['resource'], {'sameAs': url})
            if response is not None:
                return response[data_relation['field']]
                #identifier =
                #return str(identifier) if type(identifier) is ObjectId else identifier
            else:
                raise SchemaError('There is no {} with the field {} "{}".'.format(data_relation['resource'], field_name,
                                                                                  url))
        else:
            return url



from ereuse_devicehub.utils import Naming
from flask import current_app as app


def add_components(events: dict):
    """
    Adds the new devices to the materialized attribute 'components' of the parent device.
    """
    for event in events:
        app.data.driver.db['devices'].update(
            {'_id': event['device']},
            {'$addToSet': {'components': {'$each': event['components']}}}
        )


def delete_components(resource_name: str, add: dict):
    if add.get('@type') == 'devices:Add':
        app.data.driver.db['devices'].update(
            {'_id': add['device']},
            {'$pull': {'components': {'$in': add['components']}}}
        )



from ereuse_devicehub.resources.device.benchmark_settings import BenchmarkHardDrive, BenchmarkProcessor, Benchmark
from ereuse_devicehub.resources.device.schema import Device
from ereuse_devicehub.resources.device.settings import DeviceSettings, DeviceSubSettings


class Component(Device):
    interface = {
        'type': 'string',
        'teaser': False,
        'sink': -1
    }
    parent = {
        'type': 'string',
        'data_relation': {
            'resource': 'devices',
            'field': '_id',
            'embeddable': True
        }
    }

    @classmethod
    def subclasses_fields(cls):
        from ereuse_devicehub.resources.event.device import DeviceEventDomain
        global_types = super(Component, cls).subclasses_fields()
        try:
            global_types['size']['type'] = global_types['speed']['type'] = 'number'
            events = {DeviceEventDomain.new_type(x) for x in ('EraseSectors', 'EraseBasic')}
            global_types['erasure']['schema']['@type']['allowed'] = events
            union_of_benchmarks = BenchmarkHardDrive()
            union_of_benchmarks.update(BenchmarkProcessor())
            union_of_benchmarks['@type']['allowed'] = set(Benchmark.TYPES)
            global_types['benchmark']['schema'] = union_of_benchmarks
        except KeyError:
            pass
        return global_types


class ComponentSettings(DeviceSettings):
    _schema = Component
    resource_methods = []
    item_methods = []


class ComponentSubSettings(DeviceSubSettings):
    _schema = False



import copy

from ereuse_devicehub.resources.account.settings import unregistered_user
from ereuse_devicehub.resources.event.device.settings import components, EventWithDevices, \
    EventSubSettingsMultipleDevices


class Allocate(EventWithDevices):
    to = {
        'type': ['objectid', 'dict', 'string'],  # We should not add string but it does not work otherwise...
        'data_relation': {
            'resource': 'accounts',
            'field': '_id',
            'embeddable': True,
        },
        'schema': unregistered_user,
        'get_from_data_relation_or_create': 'email',
        'required': True,
        'sink': 2
    }
    toOrganization = {  # Materialization of the organization that, by the time of the allocation, the user worked in
        'type': 'string',
        'readonly': True
    }
    components = copy.deepcopy(components)


Allocate.components['readonly'] = True


class AllocateSettings(EventSubSettingsMultipleDevices):
    _schema = Allocate
    fa = 'fa-hand-o-right'
    sink = -5
    extra_response_fields = EventSubSettingsMultipleDevices.extra_response_fields + ['to']
    short_description = 'Assign the devices to someone, so that person \'owns\' the device'

# Receiver OR ReceiverEmail. We need to hook this in a required field so it is always executed
# And @type is an always required field so we can happily hook on it



"""
    Default API settings. Override them in a separate file, exactly as done in python-eve.

    These settings overrides eve's default configuration, and adds new ones.
"""
from ereuse_devicehub.resources.event.device import DeviceEventDomain
from ereuse_devicehub.utils import Naming

CLIENT = None
"""
    The base URL of DeviceHubClient or any other compatible client. The server redirects browsers to this URL.
    Do not add final trailing slash.
    Example: https://www.devicetag.io
"""

GRD = True  # GRD is part of logger, so in order for using GRD you need Logger
"""Activates the GRD Submitter."""
GRD_DEBUG = False
"""Do not really send data to GRD, just print it on screen and log it in a file."""
BASE_PATH_SHOWN_TO_GRD = None
"""
    The base URL shown to GRD when submitting data. It is supposed to be the base URL that this DeviceHub listents on.
    Add final trailing slash.
    Example: https://api.devicetag.io/
"""

# Databases
DATABASES = 'db1',
"""Sets the name of the databases, as appear in the URL. This is the 'mongoDB prefix of eve."""
DB1_DBNAME = 'dh_db1'  # The value is the name used in mongoDB for the database.
"""Sets the name of the database, as it is stored in mongoDB. Each database set in 'DATABASES' needs to have this variable."""
MONGO_DBNAME = 'dh__accounts'
"""
    Name of the central DB used only to store resources set with
    :attr:`app.resources.ResourceSettings.use_default_database
"""
RESOURCES_NOT_USING_DATABASES = ['schema']
"""List of any special resources that do not use any database, for example eve's schema endpoint."""
RESOURCES_CHANGING_NUMBER = {'device', 'event', 'account', 'place', 'erase', 'project'}
"""
    List of resources that change form singular and plural. Write it in the resource singular form.
    See :class:`app.utils.Naming`
"""

# Submitter settings
SUBMITTER_ACCOUNT = {
    'email': 'logger@ereuse.org',
    'password': '43fa22kaxl0',
}
"""The account Logger module uses to interact with DeviceHub. The account is created if it doesn't exist."""
GRD_ACCOUNT = {
    'username': None,
    'password': None
}
"""The account DeviceHub uses to send data to GRD. eReuse.org provides this account on demand."""

# Other python-eve and flask settings, no need to change them
X_HEADERS = ['Content-Type', 'Authorization']
X_EXPOSE_HEADERS = ['Authorization']
MONGO_QUERY_BLACKLIST = ['$where']
GRD_DOMAIN = 'https://sandbox.ereuse.org'
THREADS_PER_PAGE = 2
BULK_ENABLED = False  # Some events cannot work with it todo fix
CSRF_ENABLED = True
IF_MATCH = False  # We do not need concurrency control for PUT (if true, we need to provide an etag (include it in x-headers!))
XML = False  # Will probably cause bugs
# 12 hours of cache for /schema, resources have their own cache in their ResourceSettings
# See https://www.mnot.net/cache_docs/ for more info with cache
CACHE_CONTROL = 'private, max-age=' + str(60 * 60 * 12)
PAGINATION_DEFAULT = 30
PAGINATION_LIMIT = 100
DATE_FORMAT = '%Y-%m-%dT%H:%M:%S'
SCHEMA_ENDPOINT = 'schema'
VALIDATION_ERROR_AS_LIST = True  # We always use list to show errors
ITEM_CACHE = 120  # We differentiate from Resource cache (Cache setting from Eve) from Item cache

# Role settings
from ereuse_devicehub.resources.account.role import Role

ALLOWED_WRITE_ROLES = {Role.AMATEUR}
ALLOWED_ITEM_WRITE_ROLES = {Role.AMATEUR}
ALLOWED_READ_ROLES = {Role.BASIC}
ALLOWED_ITEM_READ_ROLES = {Role.BASIC}

# GRD Settings, do not change them
_events_in_grd = ('Deallocate', 'Migrate', 'Allocate', 'Receive',
                  'Remove', 'Add', 'Register', 'Locate', 'UsageProof', 'Recycle')
EVENTS_IN_GRD = [Naming.resource(DeviceEventDomain.new_type(event)) for event in _events_in_grd]
# Generation of the API (DOMAIN)
from ereuse_devicehub.resources.device.settings import DeviceSettings
from ereuse_devicehub.resources.account.settings import AccountSettings
from ereuse_devicehub.resources.event.settings import EventSettings
from ereuse_devicehub.resources.place.settings import PlaceSettings

DOMAIN = {
    'devices': DeviceSettings,
    'events': EventSettings,
    'accounts': AccountSettings,
    'places': PlaceSettings
}



from flask import current_app as app


def remove_components(events: dict):
    """
    Removes the components from the materialized attribute 'components' of the parent device.
    """
    for event in events:
        app.data.driver.db['devices'].update(
            {'_id': event['device']},
            {'$pull': {'components': {'$in': event['components']}}}
        )



from ..schema import UnitCodes, Thing


class Product(Thing):
    model = {
        'type': 'string',
        'sink': 4
    }
    weight = {
        'type': 'float',
        'unitCode': UnitCodes.kgm,
        'sink': -1,
        'teaser': False
    }
    width = {
        'type': 'float',
        'unitCode': UnitCodes.m,
        'sink': -1,
        'teaser': False
    }
    height = {
        'type': 'float',
        'unitCode': UnitCodes.m,
        'sink': -1,
        'teaser': False
    }
    manufacturer = {
        'type': 'string',
        'sink': 4
    }
    productId = {
        'type': 'string',
        'sink': 3,
        'teaser': False
    }


class IndividualProduct(Product):
    serialNumber = {
        'type': 'string',
        'sink': 4
    }


class Device(IndividualProduct):
    """Class Device"""
    _id = {
        'type': 'string',
        'unique': True,
        'device_id': True,
        'sink': 4,
        'teaser': False
        # ALLOWED_WRITE_ROLES: Role.SUPERUSER  # For recovery purposes
    }
    hid = {
        'type': 'hid',
        'sink': 5,
        'teaser': False
    }
    pid = {
        'type': 'string',
        'unique': True,
        'sink': 5
    }
    isUidSecured = {
        'type': 'boolean',
        'default': True,
        'teaser': False
    }
    labelId = {
        'type': 'string',  # Materialized label of the last snapshot
        'sink': 5
    }
    components = {
        'type': 'list',
        'schema': {
            'type': 'string',
            'data_relation': {
                'resource': 'devices',
                'embeddable': True,
                'field': '_id'
            }
        },
        'sink': 1,
        'default': []
    }
    place = {
        'type': 'objectid',
        'data_relation': {
            'resource': 'places',
            'embeddable': True,
            'field': '_id'
        },
        'materialized': True,
        'sink': 2
    }
    owners = {
        'type': 'list',
        'schema': {
            'type': 'objectid',
            'data_relation': {
                'resource': 'accounts',
                'embeddable': True,
                'field': '_id'
            }
        },
        'materialized': True,
        'sink': 2
    }
    public = {
        'type': 'boolean',
        'default': False
    }
    events = {
        'type': 'list',
        'schema': {
            'type': 'dict'
        },
        'materialized': True
    }



from ereuse_devicehub.resources.device.component.settings import Component, ComponentSubSettings
from ereuse_devicehub.resources.schema import UnitCodes


class NetworkAdapter(Component):
    speed = {
        'type': 'float',
        'unitCode': UnitCodes.mbps,
        'sink': 1
    }


# network_adapter['serialNumber']['regex'] = '^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$' todo change test inputs so we can use it again

class NetworkAdapterSettings(ComponentSubSettings):
    _schema = NetworkAdapter



from flask import current_app

from .projection import project


def project_item(resource: str, item: dict):
    project(resource, item)


def project_resource(resource: str, response: dict):
    """
    The same as project_item, but for a whole resource (list of items)
    :param resource:
    :param response:
    :return:
    """
    for item in response['_items']:
        project(resource, item)


def authorize_public(resource_name: str, item: dict):
    """
    Check if the item endpoint needs to be public, if there is a need for.

    :param resource_name:
    :param item:
    """
    if current_app.auth.needs_to_be_public():
        if not item.get('public', False):
            current_app.auth.authenticate()


def deny_public(resource_name: str, response: dict):
    """
    Hack to avoid resource endpoints to be publicly exposed
    :param resource_name:
    :param response:
    """
    if current_app.auth.needs_to_be_public():
        current_app.auth.authenticate()



class Translator:
    """
        Translates (or transforms) the structure of a resource to adequate it to another agent.

        Translation is done by specifying a translation dictionary, which is used to change from
        an original resource to the final one.
        Every field of the translation dict contains a method used to transform the original value from
        DeviceHub to the final one. Translator comes with some transformers, and subclass it to add more.

        Translation dictionaries are as follows:
        For generic translation dict: ['final_field_name'] = (transformer_method, 'original field name')
        For specific translation dicts: ['resource type name']['final field name'] = (transformer_method, 'original field name')
        Where 'final field name' is the name of the field in the agent, 'original field name' is the field name
        in DeviceHub (only add it if final name and original name differ), transformer_method is one of the transformer
        methods in Translator, and 'resource type name' e.g. devices:Register.
        See :func `GRDTranslator.__init__`: for an example.
    """
    def __init__(self, config, generic_resource: dict, translation_dict: dict):
        """
        Configures the translator. Once done, you can translate many resources as you want with :func `translate`:.
        :param config:
        :param generic_resource: Generic translation dictionary shared among resources.
        :param translation_dict: Specific translation dictionary divided per resource.
        """
        self.config = config
        self.generic = generic_resource
        self.dictionary = translation_dict
        self.database = None

    def translate(self, database: str, resource: dict) -> list:
        """
        Translates a resource.
        :param database: The database in DeviceHub (i.e. db1)
        :param resource: The resource to translate
        :return: A list of tuples, containing 1. the translated resource, 2. the original resource
        """
        self.database = database
        return [(self._translate(resource), resource)]

    def _translate(self, resource: dict) -> dict:
        """
        Translates a resource. This method carries the actual translation.
        :param resource:
        :return: The translated resource
        """
        translated = dict()
        for final_name, (method, *original_name) in dict(self.generic, **self.dictionary[resource['@type']]).items():
            value = resource.get(original_name[0] if len(original_name) > 0 else final_name)
            if value is not None:
                translated[final_name] = method(value)
        return translated

    # Transformers
    def url(self, resource_name: str):
        """Obtains an url from the resource identifier.
        :param resource_name: full resource-name with the prefix, if needed it.
        """
        def url(resource_or_identifier):
            try:
                return self._get_resource_url(resource_or_identifier['_id'], resource_name)
            except TypeError:
                return self._get_resource_url(resource_or_identifier, resource_name)
        return url

    def for_all(self, method):
        """Executes a transformer method for each value and returns a list of results"""
        def _loop(values: list) -> list:
            return [method(value) for value in values]
        return _loop

    def identity(self, value):
        """Returns the same value."""
        return value

    def device(self, device: dict) -> dict:
        """Obtains the device"""
        return device

    def hid_or_url(self, device: dict):
        """Returns HID if exists, or the URL of the device otherwise."""
        return device.get('hid', self._get_resource_url(device['_id'], 'devices'))

    # Helpers
    def _get_resource_url(self, identifier, resource_name: str):
        if self.config['DOMAIN'][resource_name]['use_default_database']:
            url = '{}/{}'.format(resource_name, identifier)
        else:
            url = '{}/{}/{}'.format(self.database, resource_name, identifier)
        return self._parse_url(url)

    def _parse_url(self, url):
        if self.config['URL_PREFIX']:
            url = '{}/{}'.format(self.config['URL_PREFIX'], url)
        return self.config['BASE_PATH_SHOWN_TO_GRD'] + '/' + url

    def inner_field(self, field: str):
        """Gets a field that is one level nested in a dict"""
        def _inner_field(value: dict):
            return value[field]
        return _inner_field



import os

import requests

from ereuse_devicehub.default_settings import GRD_DOMAIN
from ereuse_devicehub.security.request_auth import Auth
from ereuse_devicehub.tests import TestStandard


class Dummy:
    def __init__(self, domain: str, database: str, email: str, password: str):
        self.domain = domain
        self.database = database
        self.email = email
        self.password = password
        self.auth = Auth(self.domain, self.email, self.password)

    def _execute(self):
        response = requests.get('{}/api/places'.format(GRD_DOMAIN), auth=GrdAuth())
        response.raise_for_status()
        places = response.json()['results']
        for place in places:
            place['type'] = 'CollectionPoint'
            place['@type'] = 'Place'
            del place['url']
            response = requests.post('{}/{}/{}'.format(self.domain, self.database, 'places'), json=place,
                                     auth=self.auth)
            response.raise_for_status()

    def _execute2(self):
        computers_id = []
        for device in self.get_devices_from_files():
            r = requests.post('{}/{}/{}'.format(self.domain, self.database, 'snapshot'), json=device, auth=self.auth)
            r.raise_for_status()
            computers_id.append(r.json()['_id'])

    @staticmethod
    def get_devices_from_files():
        this_directory = os.path.dirname(os.path.realpath(__file__))
        file_directory = os.path.join(this_directory, '..', 'test_events', 'test_snapshot', 'resources', '2015-12-09')
        for filename in os.listdir(file_directory):
            if 'json' in filename:
                yield TestStandard.get_json_from_file(filename, file_directory)
